import dtflw.databricks as db
import dtflw.com as com


class Argument:
    """
    An argument passed to a notebook.
    """

    NAME_SUFFIX = ""

    def __init__(self, name, value):
        if name is None or len(name) == 0:
            raise ValueError("'name' cannot be None nor an empty string.")
        self.__name = name

        if value is None or type(value) != str:
            raise ValueError("'value' cannot be None and must be a string.")
        self.__value = value

    @property
    def name(self) -> str:
        return self.__name

    @property
    def value(self) -> str:
        """
        Value is always of type string.
        """
        return self.__value

    @property
    def has_value(self):
        """
        Always True for Arg.
        """
        return True

    def __repr__(self):
        return self.value

    @classmethod
    def _update_name_and_value(cls, name, value):
        widget_name = cls.get_full_name(name)
        # If a widget already exists then the line below has not effect.
        dbutils = db.get_dbutils()
        dbutils.widgets.text(widget_name, str(value), widget_name)

        return name, dbutils.widgets.get(widget_name)

    @classmethod
    def create(cls, *arguments):
        """
        Returns a dict[str, Argument | Input | Output].
        """
        names_and_values = {}
        if len(arguments) == 1 and isinstance(arguments[0], dict):
            names_and_values = arguments[0]
        else:
            names_and_values = {name: "" for name in arguments}
            shared_values = cls._get_shared_values()

            names_and_values = {**names_and_values, ** shared_values}

        return {
            name: cls(*cls._update_name_and_value(name, value))
            for name, value
            in names_and_values.items()
        }

    @classmethod
    def _get_shared_values(cls):
        return com.NotebooksChannel().get_args(
            db.get_this_notebook_abs_path()
        )

    @classmethod
    def get_full_name(cls, name: str):
        return f"{name}{cls.NAME_SUFFIX}"


class Input(Argument):
    """
    An input table required by a notebook.
    """
    NAME_SUFFIX = "_in"

    @property
    def has_value(self):
        """
        Returns False if value is an empty string. Otherwise True.
        """
        return len(self.value) > 0

    @classmethod
    def _get_shared_values(cls):
        return com.NotebooksChannel().get_inputs(
            db.get_this_notebook_abs_path()
        )


class Output(Argument):
    """
    An output table promissed by a notebook.
    """
    NAME_SUFFIX = "_out"

    @property
    def has_value(self):
        """
        Returns False if value is an empty string. Otherwise True.
        """
        return len(self.value) > 0

    @classmethod
    def _get_shared_values(cls):
        return com.NotebooksChannel().get_outputs(
            db.get_this_notebook_abs_path()
        )



import unittest
from unittest.mock import patch
from dtflw.flow import Flow
from dtflw.flow_context import FlowContext
from dtflw.lazy_notebook import LazyNotebook
from dtflw.logger import DefaultLogger
from dtflw.plugin import FlowPluginBase, NotebookPluginBase


class TestFlowPlugin(FlowPluginBase):

    @property
    def action_name(self):
        return "do_new_stuff"

    def act(self, flow: Flow, name):
        """
        Greets.
        """
        return f"Hello: {name}"


class TestNotebookPlugin(NotebookPluginBase):

    @property
    def action_name(self):
        return "get_arg_value"

    def act(self, notebook: LazyNotebook, flow: Flow, arg_name: str):
        """
        Returns arg name.
        """
        return notebook.get_args()[arg_name]


class FlowTestCase(unittest.TestCase):

    def test_install_flow_plugin(self):

        # Arrange
        ctx = FlowContext(None, None, None, DefaultLogger())
        flow = Flow(ctx)

        plg = TestFlowPlugin()
        expected = plg.act(flow, "Alice")

        # Act
        flow.install(plg)
        actual = flow.do_new_stuff("Alice")

        # Assert
        self.assertEqual(expected, actual)

    def test_install_wrong_plugin_type(self):

        # Arrange
        ctx = FlowContext(None, None, None, DefaultLogger())
        flow = Flow(ctx)

        # Act/Assert
        with self.assertRaises(ValueError):
            flow.install("This is wrong plugin type")

    def test_install_notebook_plugin(self):
        # Arrange
        ctx = FlowContext(None, None, None, DefaultLogger())
        flow = Flow(ctx)

        plg = TestNotebookPlugin()
        expected = "foo"

        # Act
        flow.install(plg)

        actual = (
            flow.notebook("import_data")
                .args({"a": expected})
                .get_arg_value("a")
        )

        # Assert
        self.assertEqual(expected, actual)

    def test_install_notebook_plugin_twice_fails(self):
        # Arrange
        ctx = FlowContext(None, None, None, DefaultLogger())
        flow = Flow(ctx)

        plg = TestNotebookPlugin()

        flow.install(plg)

        # Act/Assert
        with self.assertRaisesRegex(ValueError, f"Notebook plugin {plg.action_name} is already installed."):
            flow.install(TestNotebookPlugin())

    def test_notebook(self):

        # Arrange
        nb_plg = TestNotebookPlugin()
        ctx = FlowContext(None, None, None, DefaultLogger())
        flow = Flow(ctx)
        flow.install(nb_plg)

        # Act
        nb = flow.notebook("project/nb")

        # Assert
        self.assertEqual(nb.rel_path, "project/nb")
        self.assertTrue(hasattr(nb, nb_plg.action_name))



from dtflw.flow_context import FlowContext


class InputTable():
    """
    Represents a table which is required by a notebook as an input.
    """

    def __init__(self, name: str, abs_file_path: str, ctx: FlowContext):
        self.__name = name
        self.__abs_file_path = abs_file_path
        self.__ctx = ctx

    @property
    def name(self) -> str:
        return self.__name

    @property
    def abs_file_path(self) -> str:
        """
        Returns absolute file path of the table, or None if no suitable table was found.
        """
        return self.__abs_file_path

    def needs_eval(self) -> bool:
        """
        Returns True if the table needs to be evaluated, and False otherwise.
        """
        return not self.abs_file_path or len(self.__ctx.storage.list(self.abs_file_path)) == 0

    def validate(self, strict: bool = False):
        """
        Checks if the table is valid.
        Raises an exception if not.

        Parameters
        ----
        strict: bool
            Default: False
            No used here but added for compatiblity with output_table.validate.
        """
        if self.needs_eval():
            raise Exception("Required input not found.")



from pyspark.sql import SparkSession


def run_notebook(path: str, timeout: int, arguments: dict):
    """
    Runs the current notebook.
    """
    return get_dbutils().notebook.run(path, timeout, arguments)


def get_spark_session() -> SparkSession:
    """
    Returns a SparkSession instance.
    """
    return SparkSession.builder.getOrCreate()


def get_dbutils():
    """
    Returns a DBUtils instance.
    """
    spark = get_spark_session()

    if spark.conf.get("spark.databricks.service.client.enabled") == "true":
        from pyspark.dbutils import DBUtils
        return DBUtils(spark)
    else:
        import IPython
        return IPython.get_ipython().user_ns["dbutils"]


def get_current_username() -> str:
    """
    Returns current spark username.
    """
    dbutils = get_dbutils()
    return dbutils.notebook.entry_point.getDbutils().notebook().getContext().tags().apply('user')


def get_this_notebook_abs_path() -> str:
    """
    Returns an absolute path of the current notebook in the workspace.
    """
    return get_dbutils().notebook().entry_point.getDbutils().notebook().getContext().notebookPath().get()


def get_this_notebook_abs_cwd() -> str:
    """
    Returns an absolute path to a folder of the current notebook in the workspace.
    """
    return '/'.join(get_this_notebook_abs_path().split('/')[:-1])


def get_notebook_abs_path(rel_path) -> str:
    """
    Returns an absolute path to a notebook based on its relative path.
    """
    return '/'.join([get_this_notebook_abs_cwd(), rel_path])


def get_path_relative_to_project_dir(rel_path: str) -> str:
    """
    Returns a path starting with a project's directory for a given relative path.

    Example:

    If 'rel_path' is "dir/notebook" and the function is called from '/Repos/user@a.b/project/main'
    then 'project/dir/notebook' is returned.
    """
    return "/".join(get_notebook_abs_path(rel_path).split("/")[3:])


def try_get_context_tag(key, defaut=None):
    """
    Tries to get a value from `dbutils.notebook.entry_point.getDbutils().notebook().getContext().tags()`
    by a given key. Returns default if not found.
    """
    try:
        dbutils = get_dbutils()
        return dbutils.notebook.entry_point.getDbutils().notebook().getContext().tags().apply(key)
    except Exception as e:
        if "java.util.NoSuchElementException: key not found" in str(e):
            return defaut
        else:
            raise


def is_job_interactive() -> bool:
    """
    Returns True if the current notebook is executed in an interactive job, and False otherwise.
    """
    job_type = try_get_context_tag("jobType")
    return job_type is None


def set_runtime_config_property(key: str, value: str):
    """
    Wraps `spark.conf.set`.
    """
    get_spark_session().conf.set(key, value)


def get_runtime_config_property(key: str):
    """
    Wraps `spark.conf.get`.
    """
    return get_spark_session().conf.get(key)


def runtime_config_has(key: str):
    """
    Returns True if spark.conf has a value by the given `key`.
    """
    try:
        get_runtime_config_property(key)
    except Exception as e:
        if "java.util.NoSuchElementException" in str(e):
            return False
    return True



def assert_schema_compatible(actual_dtypes, expected_dtypes, is_strict: bool = False):
    """
    Checks if two given dataframe schemas are compatible.
    IMPORTANT: types 'int' and 'bigint' are considered the same.

    Parameters
    ------------
    actual_dtypes: list[(str, str)]
        An actual schema given as a list of columns with types.
    expected_dtypes: list[(str, str)]
        An expected schema given as a list of columns with types.
    is_strict: bool = False
        If False then the actual schema is allowed to have more columns then expected one.
        If True then actual schema must have the same columns as the expected one.
    """

    if actual_dtypes is None:
        raise ValueError("actual_dtypes must not be None.")

    if expected_dtypes is None:
        raise ValueError("expected_dtypes must not be None.")

    df_dtypes = actual_dtypes
    df_columns = map(lambda dtype: dtype[0], actual_dtypes)
    expected_columns = expected_dtypes

    errors = []

    for exp_col_name, exp_col_type in expected_columns:
        act_col = next(
            (t for t in df_dtypes if t[0] == exp_col_name), None)
        if not act_col:
            errors.append(
                f"Column '{exp_col_name}' is not present in the table.")
            continue
        act_col_type = act_col[1]

        # Pass validation for column of type `int` in case of `bigint` type is expected.
        if act_col_type == 'int' and exp_col_type == 'bigint':
            act_col_type = 'bigint'

        if exp_col_type != act_col_type and exp_col_type is not None:
            errors.append(
                f"Column's '{exp_col_name}' expected type is '{exp_col_type}' but '{act_col_type}' is given.")

    if is_strict:
        expected_column_names = set(col_name for (
            col_name, col_type) in expected_columns)
        additional_columns = [
            c for c in df_columns if c not in expected_column_names]
        if additional_columns:
            errors.append(
                f"Additional columns were present in the table. Columns: {additional_columns}")

    if errors:
        raise AssertionError('\n'.join(errors))



import dtflw.databricks as db
import json


class MessageBus:
    """
    Implements a simple messaging via `spark.cong` object.
    A bus object is a JSON of the structure:

    ROOT_CHANNEL: {
        channel: {
            topic: message
        }
    }
    """
    ROOT_CHANNEL = "flow.bus"

    def share(self, channel: str, topic: str, message):
        """
        Shares a message of a given topic on a given channel. If there is already a message
        under channel-topic then it will be replaced with a given one.

        Parameters
        ----------
        channel: str
            Name of a channel.
        topic: str
            Name of a topic.
        message: object
            A message.
        """
        if channel is None or len(channel) == 0:
            raise ValueError("`channel` cannot be None nor an empty string.")

        if topic is None or len(topic) == 0:
            raise ValueError("`topic` cannot be None nor an empty string.")

        bus = {}
        if db.runtime_config_has(self.ROOT_CHANNEL):
            bus = json.loads(db.get_runtime_config_property(self.ROOT_CHANNEL))

        if channel not in bus:
            bus[channel] = {}

        bus[channel][topic] = message

        db.set_runtime_config_property(self.ROOT_CHANNEL, json.dumps(bus))

    def try_get(self, channel: str = None, topic: str = None, defualt=None):
        """
        Returns a message under a given channel-topic if it exists.
        Otherwise, returns `default`.

        If `channel` is None then returns the complete bus content.
        If `topic` is None then returns the complete `channel` content.
        """

        if not db.runtime_config_has(self.ROOT_CHANNEL):
            return defualt

        bus = json.loads(db.get_runtime_config_property(self.ROOT_CHANNEL))

        if channel is None:
            return bus

        if channel not in bus:
            return defualt

        if topic is None:
            return bus[channel]

        if topic not in bus[channel]:
            return defualt

        return bus[channel][topic]


class NotebooksChannel:
    """
    Implements a channel specific to messaging of notebooks related content.
    """

    CHANNEL = "notebooks"

    def __init__(self):
        self.__bus = MessageBus()

    def __share(self, notebook_abs_path, sub_topic: str, message):

        existing_message = self.__get(notebook_abs_path)
        updated_message = {**existing_message, **{sub_topic: message}}

        self.__bus.share(self.CHANNEL, notebook_abs_path, updated_message)

    def share_args(self, notebook_abs_path: str, args):
        self.__share(notebook_abs_path, "args", args)

    def share_inputs(self, notebook_abs_path: str, inputs):
        self.__share(notebook_abs_path, "inputs", inputs)

    def share_outputs(self, notebook_abs_path: str, outputs):
        self.__share(notebook_abs_path, "outputs", outputs)

    def __get(self, notebook_abs_path: str, sub_topic: str = None):
        message = self.__bus.try_get(self.CHANNEL, notebook_abs_path, {})
        if sub_topic is None:
            return message
        return message[sub_topic] if sub_topic in message else {}

    def get_args(self, notebook_abs_path: str):
        return self.__get(notebook_abs_path, "args")

    def get_inputs(self, notebook_abs_path: str):
        return self.__get(notebook_abs_path, "inputs")

    def get_outputs(self, notebook_abs_path: str):
        return self.__get(notebook_abs_path, "outputs")



from abc import ABC, abstractmethod


class DisplayBase(ABC):
    """
    A base type for implementing a display service.
    Display service is used by Flow for interacting with a user in a notebook.
    """
    @abstractmethod
    def show(self, content):
        """
        Shows a given `content` object to a user.
        """
        raise NotImplemented()


class DefaultDisplay(DisplayBase):
    """
    A default display service to interact with a user in a notebook.
    """

    def show(self, content):
        """
        Shows a given `content` object to a user.
        """
        print(content)



import dtflw.assertions as A
import unittest
from ddt import ddt, data, unpack


@ddt
class AssertSchemaCompatibleTestCase(unittest.TestCase):

    @data(
        ([("foo", "int"), ("bar", "double")], [
         ("foo", "bigint"), ("bar", "double")], True),
        ([], [], True),
        ([("foo", "int"), ("bar", "double")], [("foo", "int")], False),
        ([("foo", "int"), ("bar", "double")], [], False),
        ([], [], False)
    )
    @unpack
    def test_compatile_succeeds(self, actual, expected, strict):

        try:
            A.assert_schema_compatible(actual, expected, is_strict=strict)
        except Exception as e:
            self.fail(f"Oops: {e}")

    @data(
        ([("foo", "int"), ("bar", "double")], [
         ("foo", "bigints"), ("bar", "double"), ("baz", "string")], True),
        ([("foo", "int"), ("bar", "double")], [("foo", "bigints")], True),
        ([], [("foo", "bigints")], True),
        ([("foo", "int"), ("bar", "double")], [], True),
        ([("foo", "int")], [("foo", "bool")], False),
        ([], [("foo", "bool")], False)
    )
    @unpack
    def test_compatible_fails(self, actual, expected, strict):

        with self.assertRaises(AssertionError):
            A.assert_schema_compatible(actual, expected, strict)



import unittest
from dtflw.pipeline import NotebookRun, PipelineState

class PipelineStateTestCase(unittest.TestCase):

    def test_record_run(self):
        # Arrange
        p = PipelineState()

        # Act
        r1 = NotebookRun(
            "notebook_1",
            {"arg": "val"},
            {"input": "input_path"},
            {"output": "output_path"}
        )
        p.record_run(r1)

        # Assert
        actual_run = list(p.runs)[0]

        self.assertEqual("notebook_1", actual_run.notebook_path)
        self.assertEqual({"arg": "val"}, actual_run.args)
        self.assertEqual({"input": "input_path"}, actual_run.inputs)
        self.assertEqual({"output": "output_path"}, actual_run.outputs)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description_content = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="dtflw",
    version="0.5.1",

    description="dtflw is a Python framework for building modular data pipelines based on Databricks dbutils.notebook API.",
    long_description=long_description_content,
    long_description_content_type="text/markdown",

    url="https://github.com/SoleyIo/dtflw",

    author="Soley GmbH",
    author_email="codeofconduct@soley.io",

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],

    keywords="databricks, data pipelines, etl, data engineering",

    package_dir={"": "src"},
    packages=find_packages(where="src"),

    python_requires=">=3.8",

    install_requires=[
        "ddt>=1.5.0",
        "databricks-connect",
        "setuptools"
    ],

    project_urls={
        "Bug Reports": "https://github.com/SoleyIo/dtflw/issues",
        "Source": "https://github.com/SoleyIo/dtflw"
    },
)



from __future__ import annotations
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from dtflw.databricks import get_dbutils
import fnmatch
from abc import ABC, abstractmethod


def file_exists(abs_path: str, dbutils=None) -> bool:
    """
    Returns True if a file by the given absolute path exists, and False otherwise.
    Works for Azure Storage container and DBFS.

    Parameters
    ----------
    abs_path : str
        Absolute file path.

    dbutils : DBUtils

    Returns
    -------
    True if a file exists and False otherwise.
    """
    if abs_path is None or len(abs_path) == 0:
        raise ValueError("'abs_path' cannot be None neither an empty string.")

    if dbutils is None:
        dbutils = get_dbutils()

    try:
        dbutils.fs.ls(abs_path)
        return True
    except Exception as e:
        if "java.io.FileNotFoundException" in str(e):
            return False
        else:
            raise


class FileStorageBase(ABC):
    """
    Inherit from this class to implement a specific file storage.
    The base class provides generic functionality for working with *.parquet files.
    """

    def __init__(self, spark: SparkSession, dbutils, root_dir: str = ""):
        """
        Initializes an instance.

        Parameters
        ----------
        spark: SparkSession
            A Spark session object.
        dbutils: DBUtils
            A DBUtils object.
        root_dir : str
            Root dir in the container. Default is an empty string.
        """
        self.__spark = spark
        self.__dbutils = dbutils
        self.__root_dir = root_dir

    @property
    def root_dir(self):
        """
        Root directory in the storage.
        """
        return self.__root_dir

    @property
    @abstractmethod
    def base_path(self):
        """
        Returns the base path.
        """
        pass

    def list(self, path_pattern: str, lister=None):
        """
        Returns a possibly empty list of paths that match a given pattern.
        Uses [fnmatch](https://docs.python.org/3/library/fnmatch.html).

        IMPORTANT: This function supports only a subset of 
        [pattern syntax used by a Spark reader](https://kb.databricks.com/scala/pattern-match-files-in-path.html).
        It supports:
            - [Asterisk](https://kb.databricks.com/scala/pattern-match-files-in-path.html#asterisk)
            - [Question mark](https://kb.databricks.com/scala/pattern-match-files-in-path.html#question-mark)
            - [Character class](https://kb.databricks.com/scala/pattern-match-files-in-path.html#character-class)

        Parameters
        ----------
        path_pattern: str
            A fmatch pattern.
        lister: lambda str -> list[(name:str, path:str)]
            Returns a list of (name:str, path:str) for a given path.
            Keep it None. It is used in unit test.
        """
        if path_pattern is None:
            raise ValueError("'path_pattern' cannot be None.")

        if not path_pattern.startswith(self.base_path):
            raise ValueError("'path_pattern' has a different base.")

        if lister is None:
            def lister(path): return self.__dbutils.fs.ls(path)

        def collect_matches(current_path, parts, matches):
            if len(parts) == 0:
                return

            pattern_part = parts[0]

            for info in [info for info in lister(current_path)]:
                current_name = info.name
                if current_name.endswith("/"):
                    current_name = current_name[:-1]

                if fnmatch.fnmatch(current_name, pattern_part):

                    if len(parts) == 1:
                        # Match found
                        matches.append(info.path)
                    else:
                        collect_matches(info.path, parts[1:], matches)

        parts = path_pattern[len(self.base_path) + 1:].split("/")

        matches = []
        collect_matches(self.base_path, parts, matches)
        return matches

    def is_abs_path(self, path: str) -> bool:
        """
        Returns True if a given path starts with base path.
        """
        return path is not None and path.startswith(self.base_path)

    def get_abs_path(self, rel_path: str) -> str:
        """
        Returns an absolute path to a file or a directory given by a relative path.
        Absolute path gets built as {base_path}/{rel_path}.

        Parameters
        ----------
        rel_path : str
            Relative file path.
        """
        return f"{self.base_path}/{rel_path}"

    def exists(self, path: str) -> bool:
        """
        Returns True if a given path exists in a storage and False otherwise.
        This method does not support path patterns. See `list` method for that.
        """
        if self.is_abs_path(path):
            return file_exists(path, self.__dbutils)

        return file_exists(self.get_abs_path(path), self.__dbutils)

    def get_path_with_file_extension(self, file_path):
        """
        Returns a given file path with a file extension.

        Base implementation uses `*.parquet` files by default.
        Override this method in a subclass to change the file extension.
        """
        return f"{file_path}.parquet"

    def get_path_in_root_dir(self, rel_path):
        """
        Returns a path relative to the base path with the root directory preceding.
        """
        return f"{self.__root_dir}{'/' if self.__root_dir != '' else ''}{rel_path}"

    def read_table(self, path) -> DataFrame:
        """
        Reads a table and returns a DataFrame.

        Base implementation works with `*.parquet` files by default.
        Override this method in a subclass to change to a different file format.
        """
        return self.__spark.read.parquet(path)



import unittest
from ddt import ddt, data
from dtflw.tables_repo import TablesRepository


@ddt
class TablesRepositoryTestCase(unittest.TestCase):

    @data("material", "unknown")
    def test_resolve_table_1(self, tbl_name_to_resolve):
        """
        Expected: `None` because either
        - there is no table published by a source notebook, but not the target one;
        - or a table name is unknown.
        """

        # Arrane
        source_nb_path = "nb_01"
        repo = TablesRepository()
        repo.publish(
            {"material": f"{source_nb_path}/material.parquet"}, source_nb_path)

        # Act
        # Mimic a cell evaluated multiple times
        for i in range(0, 10):
            actual_table = repo.resolve(tbl_name_to_resolve, source_nb_path)

            # Assert
            self.assertIsNone(actual_table)

    def test_resolve_table_2(self):
        """
        Expected: the last table published right before the table 
        published by the target notebook.
        """

        # Arrane
        repo = TablesRepository()

        expected_table = "nb_01/material.parquet"
        repo.publish({"material": expected_table}, "nb_01")
        repo.publish({"material": "nb_02/material.parquet"}, "nb_02")

        # Act
        # Mimic a cell evaluated multiple times
        for i in range(0, 10):
            actual_table = repo.resolve("material", "nb_02")

            # Assert
            self.assertEqual(actual_table, expected_table)

    def test_resolve_table_3(self):
        """
        Expected: the last table published if a target notebook
        has not published anything yet (had no run).
        """

        # Arrane
        repo = TablesRepository()

        expected_table = "nb_02/material.parquet"
        repo.publish({"material": "nb_01/material.parquet"}, "nb_01")
        repo.publish({"material": expected_table}, "nb_02")

        # Act
        # Mimic a cell evaluated multiple times
        for i in range(0, 10):
            actual_table = repo.resolve("material", "nb_03")

            # Assert
            self.assertEqual(actual_table, expected_table)

    @data("", None, "root_dir")
    def test_resolve_table_4(self, storage_root_dir):
        """
        Resolve inputs only to tables published before and not after.
        """

        # Arrange
        repo = TablesRepository()
        if storage_root_dir:
            storage_root_dir = storage_root_dir + '/'

        # Mimic a run of 'nb_01': publish an output 'material'
        repo.publish(
            {"material": f"{storage_root_dir}nb_01/material.parquet"}, "nb_01")

        # Mimic a run of 'nb_02': resolve an input and publish an output
        actual_table = repo.resolve("material", "nb_02")
        self.assertEqual(
            actual_table, f"{storage_root_dir}nb_01/material.parquet")

        repo.publish(
            {"material": f"{storage_root_dir}nb_02/material.parquet"}, "nb_02")

        # Mimic a run of 'nb_03': resolve an input and publish an output
        actual_table = repo.resolve("material", "nb_03")
        self.assertEqual(
            actual_table, f"{storage_root_dir}nb_02/material.parquet")

        repo.publish(
            {"material": f"{storage_root_dir}nb_03/material.parquet"}, "nb_03")

        # Act
        # Mimic a run of 'nb_02' again
        # Expected input resolved to "nb_01/material.parquet"
        for i in range(0, 10):
            actual_table = repo.resolve("material", "nb_02")
            # Assert
            self.assertEqual(
                actual_table, f"{storage_root_dir}nb_01/material.parquet")

    def test_resolve_table_if_trg_nb_is_none(self):
        """
        If target notebook is None then expected is the latest table
        published under a given name.
        """

        # Arrane
        nb_path = "nb_01"
        expected_table = f"{nb_path}/material.parquet"

        repo = TablesRepository()
        repo.publish({"material": expected_table}, nb_path)

        # Act
        actual_table = repo.resolve("material", trg_nb_path=None)

        # Assert
        self.assertEqual(actual_table, expected_table)



from dtflw.flow_context import FlowContext
import dtflw.assertions as A


class OutputTable():
    """
    Represents a table promised by a notebook as an output.
    """

    def __init__(self, name: str, abs_file_path: str, cols: list, ctx: FlowContext):
        self.__name = name
        self.__abs_file_path = abs_file_path
        self.__expected_columns = cols
        self.__ctx = ctx

    @property
    def name(self) -> str:
        return self.__name

    @property
    def abs_file_path(self) -> str:
        return self.__abs_file_path

    def needs_eval(self) -> bool:
        """
        Returns True if the table needs to be evaluated, and False otherwise.
        """
        return not self.__ctx.storage.exists(self.abs_file_path)

    def validate(self, strict: bool = False):
        """
        Checks if the table is valid.
        Raises an exception if not.

        Parameters
        ----
        strict: bool
            Default: False
            Do a strict validation on the table, meaning validation will also fail if more columns then expected are found.
        """
        if self.needs_eval():
            raise Exception("Expected output not found.")

        df = self.__ctx.storage.read_table(self.abs_file_path)

        A.assert_schema_compatible(
            df.dtypes,
            self.__expected_columns or [],
            strict
        )
